// PROJECT: The A* Pathfinder base Class
// FILE:   ASTAR.H
// OVERVIEW:
//    -Pretty good working Astar Pathfinder Algorithm  Class.
//    -Based on classastar(Rev 1.1) Author: pat@das-netz.de (thanks)
//    -Easy to implement, also with Your
//     favorite tiles based map but nevertheless a
//     little bit tricky to handle.
//    -Needs to be optimized or modified, perhaps a pointer to
//     a spritelist to reinitialisize the A* tilemap
//     and prevent sprite collosion.
//	   Let me hear about it.
//	   Have fun!
// Revision: 1.0
// Last Update: 15.06.99
// AUTHOR: wang China 
//

//#include <windows.h>
//#include <stdio.h>
//#include <stdlib.h>

#include "stdafx.h"
#include "astar.h"

////////////////////////////////////////////////////////////////////////////////
//			Constructor Destructor                        
AStar::AStar(int* pMap, int nMinForbid)
{
   Stack = ( STACK* )calloc(1,sizeof( STACK ));
   bIsPath = FALSE;
   OPEN = NULL;
   CLOSED = NULL;
   PATH = NULL;
   
   // Sets the A* Tilemap[] append on pMap from other file
   InitAstarTileMap(pMap, nMinForbid);
}

AStar::~AStar()
{
   FreeNodes();
   free(Stack);
   if ( TileMap )
		delete [] TileMap;
}

////////////////////////////////////////////////////////////////////////////////
//	Public Member Functions                      
//create TileMap[] for AStar (0 : free ; 1 : forbidden)
void AStar::InitAstarTileMap(int* pMap, int nMinForbid)
{
	COLS = MAPWIDTH + 2; 
	ROWS = MAPHEIGHT + 2; 
	TOTAL_TILES = ROWS*COLS; 
                          
	TileMap = new int[TOTAL_TILES];
	int index = COLS+1; // A* map index
	int MapIndex = 0;

	BoundaryTiles(); // sets 1's arround the map area

	// fill up the area with current map data: 1 = obstacle;0 = free
	for (int i = 0; i < ROWS-2; i++) 
	{                                     
		for (int j = 0; j < COLS-2; j++)  
		{
			if ( pMap[MapIndex] >= nMinForbid ) 
				TileMap[index] = 1;      
			MapIndex++;                 
			index++;
		}
		index += 2; 
	}
}

////////////////////////////////////////////////////////////////////////////////
//find path from sx,sy to dx,dy
BOOL AStar::NewPath(int sx,int sy, int dx,int dy)
{
	//set points to center ////-
	sx = (sx / TILESIZE)*TILESIZE + TILESIZE / 2;
	sy = (sy / TILESIZE)*TILESIZE + TILESIZE / 2;
	dx = (dx / TILESIZE)*TILESIZE + TILESIZE / 2;
	dy = (dy / TILESIZE)*TILESIZE + TILESIZE / 2;

	//check at first
	if ( FreeTile(dx,dy)&&FreeTile(sx,sy) && 
	   (TileNum(sx,sy)!=TileNum(dx,dy)) )
	{
		FreeNodes(); 
   		FindPath(sx,sy,dx,dy);
   		return (bIsPath=TRUE);
	}
	else
  		return (bIsPath=FALSE);
}

////////////////////////////////////////////////////////////////////////////////
//check if sprite reach the goal (PATH->Parent == NULL || bIsPath == FALSE)
BOOL AStar::ReachedGoal(void) 
{                                      
	if ( !bIsPath ) 
		return TRUE;  // this looks a little bit strange
	if ( PATH->Parent != NULL )// but prevents illegal calls of 
   		return FALSE;          // ::PathNextNode() or ::PathGet*()
	else
   		return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
//get the tile number of TileMap from x,y
int AStar::TileNum(int x, int y)
{
	return((y>>SHIFT)*COLS + (x>>SHIFT)+COLS+1); 
}

////////////////////////////////////////////////////////////////////////////////
//check if tile(x,y) is free (1 : free ; 0 : forbid)
int AStar::FreeTile(int x, int y)
{         
	return(!TileMap[(y>>SHIFT)*COLS + (x>>SHIFT)+COLS+1]);
}

////////////////////////////////////////////////////////////////////////////////
//	Private Member Functions                      
//free node lists of open and closed
void AStar::FreeNodes(void)
{
  	NODE *Node, *OldNode;

	if ( OPEN != NULL )
	{
   		Node = OPEN->NextNode;
   		while ( Node != NULL )
		{
      		OldNode = Node;
      		Node = Node->NextNode;
			free(OldNode);
		}
	}

	if ( CLOSED != NULL )
	{
   		Node = CLOSED->NextNode;
   		while ( Node != NULL )
		{
			OldNode = Node;
      		Node = Node->NextNode;
     		free(OldNode);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// This function places 1 around the boundary
// so that we don't go there. Also we don't
// have to treat the boundaries as a special case
void AStar::BoundaryTiles(void) 
{                             
   int c, index = 0;         

   for (c = 0; c < TOTAL_TILES; c++)      // zero out tiles.
       TileMap[c] = 0;

   for (c = 0; c < COLS; c++)         // place 1's at boundaries.
   	 TileMap[c] = 1;

   for (c = 1;c < ROWS-1; c++)
   {
      index += COLS;
      TileMap[index] = 1;
      TileMap[index+COLS-1] = 1;
   }
   index += COLS;

   for (c = index; c < TOTAL_TILES; c++)
       TileMap[c]=1;
}

////////////////////////////////////////////////////////////////////////////////
//	A* Algorithm                      
//get best path nodes in PATH node list
void AStar::FindPath(int sx, int sy, int dx, int dy)
{
	NODE *Node, *BestNode;
   int TileNumDest;

   TileNumDest = TileNum(sx, sy);

   OPEN=( NODE* )calloc(1,sizeof( NODE ));
   CLOSED=( NODE* )calloc(1,sizeof( NODE ));

   Node=( NODE* )calloc(1,sizeof( NODE ));

   Node->g = 0;
   Node->h = (dx-sx)*(dx-sx) + (dy-sy)*(dy-sy);  // should really use sqrt().
   Node->f = Node->g + Node->h;
   Node->NodeNum = TileNum(dx, dy);
   Node->x = dx;
   Node->y = dy;

   OPEN->NextNode = Node; // make Open List point to first node
   for (;;)
   {
		BestNode = ReturnBestNode();
		if (BestNode->NodeNum == TileNumDest) 
	   		break;	// if we've found the end, break and finish
		GenerateSuccessors(BestNode,sx,sy);
   }
   PATH = BestNode; //set PATH point to first best node
}

AStar::NODE *AStar::ReturnBestNode(void)
{
	NODE *tmp;
	char message[128];

	if ( OPEN->NextNode == NULL )
	{
		sprintf(message,"No more nodes on OPEN: Perhaps tilenum destination not reachable!");
		MessageBox(0, message, "Error A* Pathfinder", MB_OK | MB_ICONERROR);
		PostQuitMessage(0);
	}

	//Pick Node with lowest f, in this case it's the first node in list
	// because we sort the OPEN list wrt lowest f. Call it BESTNODE.

	//first: pop the first node(best node) of OPEN
	tmp = OPEN->NextNode;   
	OPEN->NextNode = tmp->NextNode; // Make OPEN point to nextnode or NULL.

	//next: push it(best node) at the first node of CLOSED
	tmp->NextNode = CLOSED->NextNode; 
	CLOSED->NextNode = tmp;

	return(tmp);
}

void AStar::GenerateSuccessors(NODE *BestNode, int dx, int dy)
{
   int x, y;

		    // Upper-Left
   if ( FreeTile(x=BestNode->x-TILESIZE, y=BestNode->y-TILESIZE) )
      GenerateSucc(BestNode,x,y,dx,dy);
		    // Upper
   if ( FreeTile(x=BestNode->x, y=BestNode->y-TILESIZE) )
      GenerateSucc(BestNode,x,y,dx,dy);
		    // Upper-Right
   if ( FreeTile(x=BestNode->x+TILESIZE, y=BestNode->y-TILESIZE) )
      GenerateSucc(BestNode,x,y,dx,dy);
		    // Right
   if ( FreeTile(x=BestNode->x+TILESIZE, y=BestNode->y) )
      GenerateSucc(BestNode,x,y,dx,dy);
		    // Lower-Right
   if ( FreeTile(x=BestNode->x+TILESIZE, y=BestNode->y+TILESIZE) )
      GenerateSucc(BestNode,x,y,dx,dy);
		    // Lower
   if ( FreeTile(x=BestNode->x, y=BestNode->y+TILESIZE) )
      GenerateSucc(BestNode,x,y,dx,dy);
		    // Lower-Left
   if ( FreeTile(x=BestNode->x-TILESIZE, y=BestNode->y+TILESIZE) )
      GenerateSucc(BestNode,x,y,dx,dy);
		    // Left
   if ( FreeTile(x=BestNode->x-TILESIZE, y=BestNode->y) )
      GenerateSucc(BestNode,x,y,dx,dy);
}

void AStar::GenerateSucc(NODE *BestNode,int x, int y, int dx, int dy)
{
   int g, TileNumS, c = 0;
   NODE *Old, *Successor;

   //g(dest)=g(BestNode)+cost of getting from BestNode to dest node
   g = BestNode->g+1;

   TileNumS = TileNum(x,y);  // identification purposes

   if ( (Old=CheckOPEN(TileNumS)) != NULL ) //if in OPEN list
   {
		for( c = 0; c < 8; c++)
	   		if( BestNode->Child[c] == NULL ) 
	   			break;
		BestNode->Child[c] = Old; //Add Old to the list of BestNode's Children

		 //if new g value < Old's then reset Old's parent to point to BestNode
		if ( g < Old->g ) 
		{
			Old->Parent = BestNode;
	   		Old->g = g;
	   		Old->f = g + Old->h;
		}
   }
   else if ( (Old=CheckCLOSED(TileNumS)) != NULL ) // if in closed list
   {
		for( c = 0; c< 8; c++)
			if ( BestNode->Child[c] == NULL ) 
	   			break;
		BestNode->Child[c] = Old;// Add Old to the list of BestNode's Children 

		// if new g value is < Old's then reset Old's parent to point to BestNode
		if ( g < Old->g )  
		{
	   		Old->Parent = BestNode;
	   		Old->g = g;
	   		Old->f = g + Old->h;
	   		PropagateDown(Old); // Since we changed the g value of Old, we need
					   	        // to propagate this new value downwards, i.e.
					 	        // do a Depth-First traversal of the tree!
		}
   }
   else
   {
		Successor = ( NODE* )calloc(1,sizeof( NODE ));

		Successor->Parent = BestNode;
		Successor->g = g;
		Successor->h = (x-dx)*(x-dx) + (y-dy)*(y-dy);  
		Successor->f = g + Successor->h;     
		Successor->x = x;                
		Successor->y = y;
		Successor->NodeNum = TileNumS;
		Insert(Successor); // Insert Successor on OPEN list wrt f
		for( c =0; c < 8; c++)
			if ( BestNode->Child[c] == NULL ) 
	  		    break;
		BestNode->Child[c] = Successor;
   }
}

AStar::NODE *AStar::CheckOPEN(int tilenum)
{
   NODE *tmp;

   tmp = OPEN->NextNode;
   while ( tmp != NULL )
   {
		if ( tmp->NodeNum == tilenum )
	   		return (tmp);
		else
	  		tmp = tmp->NextNode;
   }
   return(NULL);
}

AStar::NODE *AStar::CheckCLOSED(int tilenum)
{
   NODE *tmp;

   tmp = CLOSED->NextNode;

   while ( tmp != NULL )
   {
		if ( tmp->NodeNum == tilenum )
      		return(tmp);
		else
	   		tmp = tmp->NextNode;
   }
   return(NULL);
}

void AStar::Insert(NODE *Successor)
{
   NODE *tmp1, *tmp2;
   int f;

   if ( OPEN->NextNode == NULL )
   {
		OPEN->NextNode = Successor;
		return;
   }

   // insert into OPEN successor wrt f
   f = Successor->f;
   tmp1 = OPEN;
   tmp2 = OPEN->NextNode;

   //find the place in open list
   while ( (tmp2 != NULL) && (tmp2->f < f) )
   {
   		tmp1 = tmp2;
		tmp2 = tmp2->NextNode;
   }
   //if finded then insert
   Successor->NextNode = tmp2;
   tmp1->NextNode = Successor;
}

void AStar::PropagateDown(NODE *Old)
{
   int c, g;
   NODE *Child, *Father;

   g = Old->g;            // alias.
   for ( c = 0; c < 8; c++)
   {
   		if ( (Child=Old->Child[c]) == NULL ) // create alias for faster access.
			break;
		if ( g+1 < Child->g)
		{
	   		Child->g = g+1;
	   		Child->f = Child->g + Child->h;
	  	    Child->Parent = Old; // reset parent to new path.
			Push(Child); // Now the Child's branch need to be
		}     // checked out. Remember the new cost must be propagated down.
   }

   while ( Stack->NextStackPtr != NULL )
   {
		Father = Pop();
		for (c = 0; c<8; c++)
		{
	   		if ( (Child=Father->Child[c]) == NULL )  // we may stop the propagation 2 ways: either
	  			break;
			if ( Father->g+1 < Child->g ) // there are no children, or that the g value of
	  		{                          // the child is equal or better than the cost we're propagating
	      		Child->g = Father->g+1;
	    		Child->f = Child->g+Child->h;
	    		Child->Parent = Father;
	    		Push(Child);
	      }
	   }
   }
}

////////////////////////////////////////////////////////////////////////////////
//   STACK FUNCTIONS                       
void AStar::Push(NODE *Node)
{
   STACK *tmp;

   tmp =( STACK* )calloc(1,sizeof( STACK ));
   tmp->NodePtr = Node;
   tmp->NextStackPtr = Stack->NextStackPtr;
   Stack->NextStackPtr = tmp;
}

AStar::NODE *AStar::Pop(void)
{
   NODE *tmp;
   STACK *tmpSTK;

   tmpSTK = Stack->NextStackPtr;
   tmp = tmpSTK->NodePtr;

   Stack->NextStackPtr = tmpSTK->NextStackPtr;
   free(tmpSTK);
   return(tmp);
}

///////////////////////////// end ///////////////////////////////////