// PROJECT: The A* Pathfinder base Class
// FILE:   ASTAR.H
// OVERVIEW:
//    -Pretty good working Astar Pathfinder Algorithm  Class.
//    -Based on classastar(Rev 1.1) Author: pat@das-netz.de (thanks)
//    -Easy to implement, also with Your
//     favorite tiles based map but nevertheless a
//     little bit tricky to handle.
//    -Needs to be optimized or modified, perhaps a pointer to
//     a spritelist to reinitialisize the A* tilemap
//     and prevent sprite collosion.
//	   Let me hear about it.
//	   Have fun!
// Revision: 1.0
// Last Update: 15.06.99
// AUTHOR: wang China 
//
#ifndef ASTAR_H
#define ASTAR_H

#define TILESIZE 16  // change this also to reflect tile size. 64x64.
#define SHIFT 4   // change this to reflect the the size.
		          // Ex. 64x64 tile equals 2^6. or a shift of 6
#define MAPWIDTH 40		////-
#define MAPHEIGHT 30	////-

class AStar
{
private:
	struct NODE // node structure
	{     
        long f, h;
	   	int g, tmpg;
		int x, y;
		int NodeNum;
   		NODE *Parent;
		NODE *Child[8]; // a node may have upto 8+(NULL) children.
		NODE *NextNode; // for filing purposes
	};
	NODE *OPEN;    // the node list pointers
	NODE *CLOSED;
	NODE *PATH; // pointer to the best path

	struct STACK // the stack structure
	{      
	   	NODE *NodePtr;
	    STACK *NextStackPtr;
	};
	STACK *Stack;

    BOOL bIsPath;   
    int  ROWS;        // tilemap data members, need to be initialisize
    int  COLS;        // with current map's width and height
    int  TOTAL_TILES; // to allocate memory for the
	int  *TileMap;    // pointer to the A* own tilemap data array

public:
   	AStar(int* pMap, int nMinForbid);
    ~AStar();
   	void InitAstarTileMap(int* pMap, int nMinForbid);                

	BOOL NewPath(int sx, int sy, int dx, int dy);  
    BOOL ReachedGoal(void); 
    void PathNextNode(void) { PATH = PATH->Parent; }
    int	 NodeGetX(void)      { return PATH->x; }
    int  NodeGetY(void)      { return PATH->y; }

    int TileNum(int x, int y); // returns tilenum
	int FreeTile(int x, int y); // returns 1 = true if we can move on it

private:

	void BoundaryTiles(void); // sets 1's around the map area.
    void FreeNodes(void);// free lists memory

    void FindPath(int sx, int sy, int dx, int dy);
	NODE *ReturnBestNode(void);
	void GenerateSuccessors(NODE *BestNode, int dx, int dy);
   	void GenerateSucc(NODE *BestNode,int x, int y, int dx, int dy);
	NODE *CheckOPEN(int tilenum);
	NODE *CheckCLOSED(int tilenum);
	void Insert(NODE *Successor);
	void PropagateDown(NODE *Old);
    void Push(NODE *Node); 
	NODE *Pop(void);
};

#endif // ASTAR_H

